from typing import Optional, Tuple

import torch
from torch import nn

from labml_nn.lstm import LSTM
from python_autocomplete.models import AutoregressiveModel


class LstmModel(AutoregressiveModel):
    def __init__(self, *,
                 n_tokens: int,
                 embedding_size: int,
                 hidden_size: int,
                 n_layers: int):
        super().__init__()

        self.embedding = nn.Embedding(n_tokens, embedding_size)
        self.lstm = LSTM(input_size=embedding_size,
                         hidden_size=hidden_size,
                         n_layers=n_layers)
        self.fc = nn.Linear(hidden_size, n_tokens)

    def __call__(self, x: torch.Tensor, state: Optional[Tuple[torch.Tensor, torch.Tensor]]):
        # shape of x is [seq, batch, feat]
        x = self.embedding(x)
        out, (hn, cn) = self.lstm(x, state)
        logits = self.fc(out)

        return logits, (hn, cn)



from typing import Any

import torch

from labml_helpers.module import Module


class AutoregressiveModel(Module):
    def __init__(self):
        super().__init__()

    def __call__(self, src: torch.Tensor, state: Any):
        pass



from heapq import heappush, heappop
from typing import Any, List, Optional

import torch
import torch.nn

from python_autocomplete.dataset import ID_CHARS
from python_autocomplete.train import StateUpdater

EPS_PROB = 1e-6
MIN_BEAM_PROB = 1e-4


class PredictionComplete:
    def __call__(self, text, token_str: str):
        raise NotImplementedError


class NextWordPredictionComplete(PredictionComplete):
    def __init__(self, rest: str, min_length: int):
        self.min_length = min_length
        self.rest = rest

    def __call__(self, text, token_str: str):
        if len(text) - len(self.rest) < self.min_length:
            return False

        prev_is_id = text[-1] in ID_CHARS
        last_is_id = token_str[-1] in ID_CHARS

        return prev_is_id != last_is_id


class NextWordNewLinePredictionComplete(PredictionComplete):
    def __init__(self, rest: str, min_length: int):
        self.min_length = min_length
        self.rest = rest

    def __call__(self, text, token_str: str):
        if len(text) - len(self.rest) < self.min_length:
            return False

        if '\n' in token_str:
            return True

        prev_is_id = text[-1] in ID_CHARS
        last_is_id = token_str[-1] in ID_CHARS

        return prev_is_id != last_is_id


class BeamSearch:
    def __init__(self):
        pass

    def next_batch(self, prompt: torch.Tensor, state: Any):
        raise NotImplementedError

    def update(self, next_token, state, old_state):
        raise NotImplementedError


class BeamSearchSimple(BeamSearch):
    def __init__(self, *, beam_size: int, prediction_complete: PredictionComplete,
                 max_beam_size: int, rest: str,
                 state_updater: 'StateUpdater',
                 probs: Optional[List[float]],
                 is_token_by_token: bool,
                 itos: List[str]):
        super().__init__()
        self.itos = itos
        self.is_token_by_token = is_token_by_token
        self.state_updater = state_updater
        self.prediction_complete = prediction_complete
        self.max_beam_size = max_beam_size
        self.rest = rest

        if probs is None:
            probs = [1 / beam_size] * beam_size
        assert len(probs) == beam_size
        self.probs = probs

        self.result_heap = []
        self.text = [''] * beam_size
        self.beam_heap = []

    @staticmethod
    def is_substr(original, token_str):
        if not original:
            return True

        n = min(len(original), len(token_str))
        return original[:n] == token_str[:n]

    def add_prediction(self, prob: float, beam_idx: int, token_str: str, state):
        if len(self.result_heap) == self.max_beam_size:
            if self.result_heap[0][0] > prob - EPS_PROB:
                return False
            heappop(self.result_heap)

        state = self.state_updater.get_from_batch(state, beam_idx)
        text = self.text[beam_idx] + token_str
        heappush(self.result_heap, (prob, (text, state)))

        return True

    def add_prediction_before_token(self, prob: float, beam_idx: int, state):
        if len(self.result_heap) == self.max_beam_size:
            if self.result_heap[0][0] > prob - EPS_PROB:
                return False
            heappop(self.result_heap)

        state = self.state_updater.get_from_batch(state, beam_idx)
        text = self.text[beam_idx]
        heappush(self.result_heap, (prob, (text, state)))

        return True

    def add_beam(self, prob: float, beam_idx: int, token: int):
        if self.result_heap and self.result_heap[0][0] > prob - EPS_PROB:
            return False

        if prob < MIN_BEAM_PROB:
            return False

        if len(self.beam_heap) == self.max_beam_size:
            if self.beam_heap[0][0] > prob - EPS_PROB:
                return False
            heappop(self.beam_heap)

        heappush(self.beam_heap, (prob, (beam_idx, token)))

        return True

    def next_batch(self, prompt: torch.Tensor, state: Any):
        if not self.beam_heap:
            return None, None

        new_prompt = []
        new_state = []

        texts = self.text
        self.text = []
        self.probs = []

        for prob, (b, token) in self.beam_heap:
            token = prompt.new_tensor([token])
            if self.is_token_by_token:
                new_prompt.append(token)
            else:
                new_prompt.append(torch.cat((prompt[1:, b], token)))
            new_state.append(self.state_updater.get_from_batch(state, b))
            self.probs.append(prob)
            self.text.append(texts[b] + self.itos[token])

        new_prompt = torch.stack(new_prompt, dim=1)
        new_state = self.state_updater.make_batch(new_state)

        self.beam_heap = []

        return new_prompt, new_state

    def update(self, next_token, state, old_state):
        self.beam_heap = []

        for b, text in enumerate(self.text):
            text = self.text[b]
            if len(text) >= len(self.rest):
                check_rest = None
            else:
                check_rest = self.rest[len(text):]

            tokens = next_token[b]
            sort_idx = torch.argsort(tokens)

            for i in reversed(range(len(tokens))):
                token = sort_idx[i]
                token_str = self.itos[token]
                if not self.is_substr(check_rest, token_str):
                    continue

                if self.prediction_complete(text, token_str):
                    if not self.add_prediction_before_token(self.probs[b], b, old_state):
                        break
                    else:
                        break
                    # if not self.add_prediction(self.probs[b] * tokens[token].item(), b, token_str, state):
                    #     break
                if not self.add_beam(self.probs[b] * tokens[token].item(), b, token):
                    break



from functools import lru_cache
from heapq import heappush, heappop
from typing import List

from labml import lab, monit
from labml.utils.cache import cache_set
from python_autocomplete.dataset import Tokenizer
from python_autocomplete.dataset.break_words import SourceCodeTokenizer


class BPE(Tokenizer):
    def __init__(self, bpe_en_de: 'BPEEnDe', word_tokenizer):
        self.bpe = bpe_en_de
        self.word_tokenizer = word_tokenizer
        self.is_trained = True

    @property
    def n_tokens(self):
        return len(self.bpe.bpe)

    @property
    def itos(self):
        return self.bpe.bpe_itos

    @property
    def stoi(self):
        return self.bpe.bpe_stoi

    def encode(self, data: str, *, is_silent: bool = True):
        words = self.word_tokenizer.tokenize(data, is_silent=is_silent)

        res = []
        for w in monit.iterate('Encode words', words, is_silent=is_silent):
            res += self.bpe.encode(w)

        return res

    def __call__(self, data: str):
        encoded = self.encode(data)
        return [self.itos[c] for c in encoded]

    def rstrip(self, data: str):
        words = self.word_tokenizer.tokenize(data, is_silent=True)
        words = words[:-1]
        res = []
        for w in words:
            res += self.bpe.encode(w)

        return ''.join(words), res


class _BPEEncoder:
    def __init__(self, pairs):
        self.pairs = pairs
        self.codes = []
        self.next_idx = []
        self.prev_idx = []
        self.heap = []

    def encode(self, codes: List[int]):
        self.codes = codes
        self.next_idx = BPELearner.default_next_pointers(len(codes))
        self.prev_idx = BPELearner.default_prev_pointers(len(codes))
        self.heap = []

        for i in range(len(self.codes) - 1):
            self.add_pair((self.codes[i], self.codes[i + 1]), i)

        while self.heap:
            _, idx, pair = heappop(self.heap)
            self.merge(idx, pair)

        return [c for c in self.codes if c != -1]

    def merge(self, p2, pair):
        p3 = self.next_idx[p2]

        if p3 == -1 or pair[0] != self.codes[p2] or pair[1] != self.codes[p3]:
            return

        self.codes[p2] = self.pairs[pair]
        self.codes[p3] = -1
        p1 = self.prev_idx[p2]
        p4 = self.next_idx[p3]

        if p1 != -1:
            self.add_pair((self.codes[p1], self.codes[p2]), p1)
        self.next_idx[p2] = p4
        if p4 != -1:
            self.prev_idx[p4] = p2
            self.add_pair((self.codes[p2], self.codes[p4]), p2)

    def add_pair(self, pair, idx):
        if pair not in self.pairs:
            return

        heappush(self.heap, (self.pairs[pair], idx, pair))


class BPEEnDe:
    def __init__(self):
        self.char_itos = []
        self.char_stoi = {}
        self.bpe = []
        self.popular_words = {}

        self.bpe_itos = []
        self.bpe_stoi = {}
        self.pairs = {}
        self.encoder = None

    def load(self, char_itos, char_stoi, bpe):
        self.char_itos = char_itos
        self.char_stoi = char_stoi
        self.bpe = bpe

        self.calc()

    def set_popular_words(self, popular_words):
        self.popular_words = popular_words

    def calc(self):
        self.bpe_itos = self.calc_bpe_itos()
        self.bpe_stoi = {s: i for i, s in enumerate(self.bpe_itos)}
        self.pairs = {(p[0], p[1]): c for c, p in enumerate(self.bpe) if not isinstance(p, int)}

        self.encoder = _BPEEncoder(self.pairs)

    def to_char_stoi(self, w: str):
        return [self.char_stoi[c] for c in w]

    def calc_bpe_itos(self):
        itos = list(self.char_itos)
        for p1, p2 in self.bpe[len(self.char_itos):]:
            itos.append(itos[p1] + itos[p2])
        return itos

    @lru_cache(1024)
    def encode(self, word: str):
        if word in self.popular_words:
            return self.popular_words[word]

        return self.encoder.encode([self.char_stoi[c] for c in word if c in self.char_stoi])


class BPELearner:
    def __init__(self, words_list: List[str], word_freq: List[int]):
        self.words_list = words_list
        self.word_freq = word_freq

        self.heap = []
        self.heap_modified = set()
        self.char_itos = []
        self.char_stoi = {}
        self.bpe = []
        self.word_codes = []
        self.word_code_prev = []
        self.word_code_next = []

        self.counts = {}
        self.locations = {}

        self.build_vocab()
        self.build_word_arrays()
        self.collect_pairs()

    def learn(self, merges: int):
        for i in monit.iterate('BPE', merges):
            while True:
                res = self.merge_pair()
                if res is not None:
                    break

    def build_vocab(self):
        vocab = set()
        for k in self.words_list:
            for c in k:
                vocab.add(c)

        self.char_itos = list(sorted(vocab))
        self.char_stoi = {c: i for i, c in enumerate(self.char_itos)}

        self.bpe = [i for i in range(len(self.char_stoi))]

    def to_char_stoi(self, w: str):
        return [self.char_stoi[c] for c in w]

    @staticmethod
    def default_next_pointers(length: int):
        return [i + 1 for i in range(length - 1)] + [-1]

    @staticmethod
    def default_prev_pointers(length: int):
        return [i - 1 for i in range(length)]

    def build_word_arrays(self):
        self.word_codes = [self.to_char_stoi(w) for w in self.words_list]
        self.word_code_next = [self.default_next_pointers(len(w)) for w in self.word_codes]
        self.word_code_prev = [self.default_prev_pointers(len(w)) for w in self.word_codes]

    def heap_add_all(self):
        for pair in self.heap_modified:
            if pair in self.counts:
                heappush(self.heap, (-self.counts[pair], pair))

    def add_pair(self, w, i, nxt):
        pair = self.word_codes[w][i], self.word_codes[w][nxt]
        assert pair[0] != -1 and pair[1] != -1

        if pair not in self.counts:
            self.counts[pair] = 0
            self.locations[pair] = {}

        if w not in self.locations[pair]:
            self.locations[pair][w] = set()

        self.counts[pair] += self.word_freq[w]
        self.locations[pair][w].add(i)

        self.heap_modified.add(pair)

    def collect_pairs(self):
        for w, v in monit.enum('Collect pairs', self.word_codes):
            f = self.word_freq[w]

            for i in range(len(v) - 1):
                self.add_pair(w, i, i + 1)

        self.heap_add_all()

    def remove_pair(self, w, i, nxt):
        pair = self.word_codes[w][i], self.word_codes[w][nxt]
        assert pair[0] != -1 and pair[1] != -1
        if pair not in self.counts:
            return
        self.locations[pair][w].remove(i)
        self.counts[pair] -= self.word_freq[w]
        self.heap_modified.add(pair)

    def merge_pair(self):
        cnt, pair = heappop(self.heap)
        if pair not in self.counts or self.counts[pair] != -cnt:
            return None

        n = len(self.bpe)
        self.bpe.append(pair)
        del self.counts[pair]
        for w, locs in self.locations[pair].items():
            locs = list(reversed(sorted(locs)))
            prev = None
            merged = []
            for p2 in locs:
                p1 = self.word_code_prev[w][p2]
                p3 = self.word_code_next[w][p2]
                assert p3 != -1
                if p3 == prev:
                    continue
                p4 = self.word_code_next[w][p3]

                if p1 != -1:
                    self.remove_pair(w, p1, p2)
                if p4 != -1 and p4 != prev:
                    self.remove_pair(w, p3, p4)

                prev = p2
                merged.append(p2)

            for p2 in merged:
                p3 = self.word_code_next[w][p2]
                p4 = self.word_code_next[w][p3]
                self.word_codes[w][p2] = n
                self.word_codes[w][p3] = -1
                self.word_code_next[w][p3] = -1
                self.word_code_prev[w][p3] = -1
                if p4 != -1:
                    self.word_code_next[w][p2] = p4
                    self.word_code_prev[w][p4] = p2
                else:
                    self.word_code_next[w][p2] = -1

            for p2 in merged:
                p1 = self.word_code_prev[w][p2]
                p3 = self.word_code_next[w][p2]

                if p1 != -1:
                    self.add_pair(w, p1, p2)
                if p3 != -1:
                    self.add_pair(w, p2, p3)

        self.heap_add_all()

        return pair

    def bpe_itos(self):
        itos = list(self.char_itos)
        for p1, p2 in self.bpe[len(self.char_itos):]:
            itos.append(itos[p1] + itos[p2])

        return itos

    def get_length(self):
        res = 0
        for w, v in enumerate(self.word_codes):
            cnt = 0
            for idx in v:
                if idx != -1:
                    cnt += 1
            res += cnt * self.word_freq[w]

        return res


def main():
    path = lab.get_data_path() / 'train.py'

    with open(str(path), 'r') as f:
        data = f.read()

    tokenizer = SourceCodeTokenizer()
    tokenizer.collect_words(data)

    bpe = BPELearner(*tokenizer.get_words())
    bpe.learn(1000)
    print(len(bpe.bpe))
    print(bpe.bpe_itos()[len(bpe.char_itos):])
    print(len(data), bpe.get_length())

    cache_set('bpe', {
        'char_itos': bpe.char_itos,
        'char_stoi': bpe.char_stoi,
        'bpe': bpe.bpe
    })

    bpe_en_de = BPEEnDe()
    bpe_en_de.load(bpe.char_itos, bpe.char_stoi, bpe.bpe)


if __name__ == '__main__':
    main()



from labml import logger, lab, monit
from labml.logger import Text, Style
from python_autocomplete.evaluate import Predictor
from python_autocomplete.evaluate.factory import get_predictor


def generate(predictor: Predictor, text: str, completion: int):
    line_no = 1
    logs = [(f"{line_no: 4d}: ", Text.meta), (text[0], Text.subtle)]

    i = 0
    given = len(text)

    while i + 1 < given + completion:
        if len(text) > i + 1:
            c = text[i + 1]
        else:
            c, _ = predictor.get_next_token(text[:i + 1], None)

        if c == '\n':
            logger.log(logs)
            line_no += 1
            logs = [(f"{line_no: 4d}: ", Text.meta)]
        elif c != '\r':
            if len(text) > i + 1:
                logs.append(c)
            else:
                logs.append((c, [Style.bold]))

        if len(text) <= i + 1:
            text += c

        i += 1

    logger.log(logs)


def main():
    predictor = get_predictor()

    with open(str(lab.get_data_path() / 'sample.py'), 'r') as f:
        sample = f.read()
    with monit.section('Generate'):
        generate(predictor, 'import numpy as np\n', 1000)


if __name__ == '__main__':
    main()



import string
from typing import Dict, List, Tuple

ID_CHARS = set(string.ascii_letters + string.digits + '_')


class Tokenizer:
    n_tokens: int
    itos: List[str]
    stoi: Dict[str, int]
    is_trained: int

    def encode(self, data: str, *, is_silent: bool = True):
        raise NotImplementedError

    def train(self, data: str):
        pass

    def rstrip(self, data: str) -> Tuple[str, List[int]]:
        return data, self.encode(data)



import json
import threading

import torch
from flask import Flask, request, jsonify

from labml import monit
from python_autocomplete.evaluate.beam_search import NextWordPredictionComplete
from python_autocomplete.evaluate.factory import get_predictor

app = Flask('python_autocomplete')
predictor = get_predictor()
lock = threading.Lock()


@app.route('/')
def home():
    return 'Home'


@app.route('/autocomplete', methods=['POST'])
def autocomplete():
    prefix = request.json['prompt']
    if not prefix:
        return jsonify({'success': False})

    with monit.section('Predict') as s:
        acquired = lock.acquire(blocking=False)
        if acquired:
            stripped, prompt = predictor.rstrip(prefix)
            rest = prefix[len(stripped):]
            prediction_complete = NextWordPredictionComplete(rest, 15)
            prompt = torch.tensor(prompt, dtype=torch.long).unsqueeze(-1)

            predictions = predictor.get_next_word(prompt, None, rest, [1.], prediction_complete, 5)
            predictions.sort(key=lambda x: -x.prob)

            results = [pred.text[len(rest):] for pred in predictions]
            probs = [pred.prob for pred in predictions]
            lock.release()
            s.message = f'{json.dumps(prefix[-5:])} -> {json.dumps(results)}'
            return jsonify({'success': True, 'prediction': results, 'probs': probs})
        else:
            monit.fail()
            return jsonify({'success': False})


if __name__ == '__main__':
    app.run(host='0.0.0.0', port=5000, debug=False)



from typing import List, Optional

import torch
from torch import nn

from labml_nn.transformers.xl import TransformerXL
from python_autocomplete.models import AutoregressiveModel


class TransformerXLModel(AutoregressiveModel):
    def __init__(self, n_vocab: int, d_model: int, transformer: TransformerXL):
        super().__init__()
        self.src_embed = nn.Embedding(n_vocab, d_model)
        self.transformer = transformer
        self.generator = nn.Linear(d_model, n_vocab)
        self.mask_x = None
        self.mask_mem = None

    def __call__(self, x: torch.Tensor, mem: Optional[List[torch.Tensor]]):
        m_len = len(mem[0]) if mem else 0
        if self.mask_x is None or self.mask_x.shape[0] < len(x):
            from labml_nn.transformers.utils import subsequent_mask
            self.mask_x = subsequent_mask(len(x)).to(x.device)
        if self.mask_mem is None or self.mask_mem.shape[1] < m_len or self.mask_mem.shape[0] < len(x):
            self.mask_mem = self.mask_x.new_ones(len(x), m_len, 1)

        if m_len:
            mask = torch.cat((self.mask_mem[:len(x), :m_len], self.mask_x[:len(x), :len(x)]), dim=1)
        else:
            mask = self.mask_x[:len(x), :len(x)]

        x = self.src_embed(x)
        res, mem = self.transformer(x, mem, mask)
        res = self.generator(res)

        return res, mem



from typing import Any

import torch
from torch import nn

from labml_helpers.module import Module
from labml_nn.transformers import Encoder
from labml_nn.transformers.utils import subsequent_mask
from python_autocomplete.models import AutoregressiveModel


class TransformerModel(AutoregressiveModel):
    def __init__(self, n_tokens, d_model, encoder: Encoder, src_embed: Module):
        super().__init__()
        self.src_mask = None
        self.encoder = encoder
        self.src_embed = src_embed
        self.d_model = d_model
        self.fc = nn.Linear(d_model, n_tokens)

    def __call__(self, src: torch.Tensor, _: Any = None):
        if self.src_mask is None or self.src_mask.size(0) != len(src):
            self.src_mask = subsequent_mask(len(src)).to(src.device)

        src = self.src_embed(src)
        # with monit.section("transformer"):
        output = self.encoder(src, self.src_mask)
        output = self.fc(output)
        return output, None



from typing import Any, Tuple, List, NamedTuple

import torch
import torch.nn
from torch import nn

from labml import monit
from labml_helpers.module import Module
from python_autocomplete.dataset import Tokenizer
from python_autocomplete.evaluate.beam_search import PredictionComplete, BeamSearch, BeamSearchSimple
from python_autocomplete.evaluate.beam_search_lengthy import BeamSearchLengthy
from python_autocomplete.train import StateUpdater


class Prediction(NamedTuple):
    prob: float
    text: str
    state: Any


class Predictor:
    def __init__(self, model: Module, tokenizer: Tokenizer, *,
                 state_updater: StateUpdater,
                 is_token_by_token: bool):
        self.tokenizer = tokenizer
        self.is_token_by_token = is_token_by_token
        self.state_updater = state_updater
        self.model = model

        # For timing
        self.time_add = 0
        self.time_predict = 0
        self.time_check = 0

    def _get_predictions(self, prompt: torch.Tensor, state: Any) -> Tuple[torch.Tensor, Any]:
        if prompt.shape[0] == 0:
            return prompt.new_ones(prompt.shape[1], len(self.tokenizer.itos)) / len(self.tokenizer.itos), state
        prompt = prompt.to(self.model.device)

        # Get predictions
        with torch.no_grad():
            prediction, new_state = self.model(prompt, state)

        state = self.state_updater(state, new_state)
        prediction = nn.Softmax(-1)(prediction[-1])

        # Final prediction
        return prediction, state

    def get_next_word(self, prompt: torch.Tensor, state: Any, rest: str, probs: List[float],
                      prediction_complete: PredictionComplete,
                      max_beam_size: int) -> \
            List[Prediction]:
        beam = BeamSearchSimple(beam_size=prompt.shape[1],
                                prediction_complete=prediction_complete,
                                max_beam_size=max_beam_size,
                                rest=rest,
                                state_updater=self.state_updater,
                                probs=probs,
                                is_token_by_token=self.is_token_by_token,
                                itos=self.tokenizer.itos)

        for _ in range(10):
            with monit.section('Predict', is_silent=True):
                next_token, new_state = self._get_predictions(prompt, state)
            with monit.section('Beam', is_silent=True):
                beam.update(next_token, new_state, state)
                prompt, state = beam.next_batch(prompt, new_state)

            if prompt is None:
                break

        results = [Prediction(r[0], r[1][0], r[1][1]) for r in beam.result_heap]
        return results

    def rstrip(self, prompt: str) -> Tuple[str, List[int]]:
        return self.tokenizer.rstrip(prompt)



from pathlib import PurePath
from typing import Optional

import torch
from torch.utils.data import Dataset, DataLoader

from labml import lab, monit
from labml.configs import option, BaseConfigs
from labml_helpers.datasets.text import TextDataset
from python_autocomplete.dataset import Tokenizer
from python_autocomplete.dataset.bpe import BPE, SourceCodeTokenizer


class CharacterTokenizer(Tokenizer):
    def __init__(self, retrain: bool):
        from labml.utils.cache import cache_get

        self.n_tokens = cache_get('n_tokens')
        self.itos = cache_get('itos')
        self.stoi = cache_get('stoi')
        if retrain is None:
            self.is_trained = self.n_tokens and self.itos and self.stoi
        else:
            self.is_trained = not retrain

    def encode(self, data: str, *, is_silent: bool = True):
        return [self.stoi[c] for c in data if c in self.stoi]

    def train(self, data: str):
        with monit.section("Build vocabulary"):
            self.itos = list(sorted(list(set(data))))
            self.n_tokens = len(self.itos)
            self.stoi = {c: i for i, c in enumerate(self.itos)}

        from labml.utils.cache import cache_set

        cache_set(f'n_tokens', self.n_tokens)
        cache_set(f'itos', self.itos)
        cache_set(f'stoi', self.stoi)


class SourceCodeDataset:
    @staticmethod
    def load(path: PurePath):
        with open(str(path), 'r') as f:
            return f.read()

    @staticmethod
    def get_train_valid(path: PurePath, is_load_data: bool):
        if is_load_data:
            with monit.section("Load data"):
                train = TextDataset.load(path / 'train.py')
                valid = TextDataset.load(path / 'valid.py')
        else:
            train = ''
            valid = ''

        return train, valid

    def __init__(self, tokenizer: Tokenizer, train, valid):
        self.train = train
        self.valid = valid
        self.tokenizer = tokenizer

    def __repr__(self):
        return f'{len(self.train) / 1_000_000 :,.2f}M, {len(self.valid) / 1_000_000 :,.2f}'


class SourceCodeDataConfigs(BaseConfigs):
    dataset: SourceCodeDataset
    truncate_data: int = 0
    is_load_data: bool = True
    tokenizer: Tokenizer
    retrain_tokenizer: bool = True

    train_loader: DataLoader
    valid_loader: DataLoader
    is_shuffle: bool = True
    batch_size: int
    seq_len: int

    def text_to_i(self, text: str, *, is_silent: bool = True) -> torch.Tensor:
        return torch.tensor(self.tokenizer.encode(text, is_silent=is_silent), dtype=torch.long)


@option(SourceCodeDataConfigs.dataset, 'default')
def _dataset(c: SourceCodeDataConfigs):
    train, valid = SourceCodeDataset.get_train_valid(lab.get_data_path(), c.is_load_data)
    if c.truncate_data:
        train, valid = train[:c.truncate_data], valid[:c.truncate_data]
    if not c.tokenizer.is_trained:
        c.tokenizer.train(train + valid)
    return SourceCodeDataset(c.tokenizer, train, valid)


@option(SourceCodeDataConfigs.tokenizer, 'bpe')
def _bpe_tokenizer():
    from labml.utils.cache import cache_get
    from python_autocomplete.dataset.bpe import BPEEnDe
    bpe_cache = cache_get('bpe')

    if bpe_cache:
        bpe_en_de = BPEEnDe()
        bpe_en_de.load(**bpe_cache)
    else:
        raise RuntimeError('BPE not cached')

    return BPE(bpe_en_de, SourceCodeTokenizer())


@option(SourceCodeDataConfigs.tokenizer, 'char')
def _char_tokenizer(c: SourceCodeDataConfigs):
    return CharacterTokenizer(c.retrain_tokenizer)


# Data loaders
class TokenDataset(Dataset):
    def __init__(self, *,
                 data: torch.Tensor,
                 batch_size: int,
                 seq_len: int,
                 drop_last: bool = False):
        self.batch_size = batch_size
        self.seq_len = seq_len
        self.data = data
        self.n_samples = (self.data.shape[0] - 1) // self.seq_len
        if not drop_last:
            self.n_batches = (self.n_samples + batch_size - 1) // batch_size
        else:
            self.n_batches = self.n_samples // batch_size

    def __len__(self):
        return (self.data.shape[0] - 1) // self.seq_len

    def __getitem__(self, idx):
        batch = idx // self.batch_size
        batch_idx = idx % self.batch_size
        idx = batch_idx * self.n_batches + batch
        start = idx * self.seq_len
        assert start + self.seq_len + 1 <= self.data.shape[0]
        end = start + self.seq_len
        data = self.data[start: end]
        target = self.data[start + 1: end + 1]
        return data, target


def transpose_batch(batch):
    transposed_data = list(zip(*batch))
    src = torch.stack(transposed_data[0], 1)
    tgt = torch.stack(transposed_data[1], 1)

    return src, tgt


@option(SourceCodeDataConfigs.train_loader)
def _train_loader(c: SourceCodeDataConfigs):
    return DataLoader(TokenDataset(data=c.text_to_i(c.dataset.train, is_silent=False),
                                   batch_size=c.batch_size,
                                   seq_len=c.seq_len,
                                   drop_last=True),
                      batch_size=c.batch_size,
                      collate_fn=transpose_batch,
                      shuffle=c.is_shuffle,
                      drop_last=True)


@option(SourceCodeDataConfigs.valid_loader)
def _valid_loader(c: SourceCodeDataConfigs):
    return DataLoader(TokenDataset(data=c.text_to_i(c.dataset.valid, is_silent=False),
                                   batch_size=c.batch_size,
                                   seq_len=c.seq_len,
                                   drop_last=True),
                      batch_size=c.batch_size,
                      collate_fn=transpose_batch,
                      shuffle=c.is_shuffle,
                      drop_last=True)



import datetime

import torch.distributed
import torch.nn as nn

from labml import experiment, monit, logger, tracker
from labml.configs import option
from labml.logger import Text, inspect
from labml_helpers.train_valid import hook_model_outputs, BatchIndex
from python_autocomplete.train import Configs as Configs_


class Configs(Configs_):
    ddp_model: nn.parallel.DistributedDataParallel

    def init(self):
        tracker.set_queue("loss.*", 20, True)
        tracker.set_scalar("accuracy.*", True)
        hook_model_outputs(self.mode, self.ddp_model, 'model')
        self.state_modules = [self.accuracy]

    def step(self, batch: any, batch_idx: BatchIndex):
        data, target = batch[0].to(self.device), batch[1].to(self.device)

        if self.mode.is_train:
            tracker.add_global_step(len(data))

        with self.mode.update(is_log_activations=batch_idx.is_last):
            output, *_ = self.ddp_model(data)

        loss = self.loss_func(output, target)
        self.accuracy(output, target)
        self.accuracy.track()
        tracker.add("loss.", loss)

        if self.mode.is_train:
            loss.backward()

            self.optimizer.step()
            if batch_idx.is_last:
                tracker.add('model', self.ddp_model)
            self.optimizer.zero_grad()

        tracker.save()

    def sample(self):
        prompt = 'def train('
        log = [(prompt, Text.subtle)]
        for i in monit.iterate('Sample', 25):
            data = self.text.text_to_i(prompt).unsqueeze(-1)
            data = data.to(self.device)
            output, *_ = self.ddp_model(data)
            output = output.argmax(dim=-1).squeeze()
            prompt += '' + self.text.itos[output[-1]]
            log += [('' + self.text.itos[output[-1]], Text.value)]

        logger.log(log)


@option(Configs.ddp_model)
def ddp_model(c: Configs):
    return nn.parallel.DistributedDataParallel(c.model, device_ids=[c.device])


def main(local_rank, rank, world_size, uuid, init_method: str = 'tcp://localhost:23456'):
    with monit.section('Distributed'):
        torch.distributed.init_process_group("gloo",
                                             timeout=datetime.timedelta(seconds=30),
                                             init_method=init_method,
                                             rank=rank,
                                             world_size=world_size)
    conf = Configs()
    experiment.create(uuid=uuid,
                      name="source_code_ddp",
                      comment='lstm model')
    experiment.distributed(local_rank, world_size)
    experiment.configs(conf, {
        'model': 'transformer_model',
        'n_layers': 6,
        'batch_size': 12,
        'epochs': 32,
        'optimizer.optimizer': 'Noam',
        'optimizer.learning_rate': 1.0,
        'device.cuda_device': local_rank,
        'seq_len': 512,
        'train_loader': 'shuffled_train_loader',
        'valid_loader': 'shuffled_valid_loader'
    })
    experiment.add_pytorch_models(model=conf.ddp_model)
    with experiment.start():
        conf.run()


def _launcher():
    import os
    world_size = int(os.environ['WORLD_SIZE'])
    run_uuid = os.environ['RUN_UUID']
    local_rank = int(os.environ['LOCAL_RANK'])
    rank = int(os.environ['RANK'])
    inspect(world_size=os.environ['WORLD_SIZE'],
            run_uuid=os.environ['RUN_UUID'],
            local_rank=os.environ['LOCAL_RANK'],
            rank=os.environ['RANK'],
            master_addr=os.environ['MASTER_ADDR'],
            master_port=os.environ['MASTER_PORT'])
    main(local_rank, rank, world_size, run_uuid, 'env://')


if __name__ == '__main__':
    _launcher()



#!/usr/bin/env python

"""
Parse all files and write to a single file
"""
import re
import string
import urllib.error
import urllib.request
import zipfile
from pathlib import Path
from pathlib import PurePath
from typing import List, NamedTuple, Set
from typing import Optional

import numpy as np
from labml import lab, monit
from labml import logger
from labml.internal.util import rm_tree

PRINTABLE = set(string.printable)


class PythonFile(NamedTuple):
    relative_path: str
    project: str
    path: Path


def get_python_files():
    """
    Get list of python files and their paths inside `data/source` folder
    """

    source_path = Path(lab.get_data_path() / 'source')
    files: List[PythonFile] = []

    def _add_file(path: Path):
        """
        Add a file to the list of tiles
        """
        project = path.relative_to(source_path).parents
        relative_path = path.relative_to(source_path / project[len(project) - 3])

        files.append(PythonFile(relative_path=str(relative_path),
                                project=str(project[len(project) - 2]),
                                path=path))

    def _collect_python_files(path: Path):
        """
        Recursively collect files
        """
        for p in path.iterdir():
            if p.is_dir():
                _collect_python_files(p)
            else:
                _add_file(p)

    _collect_python_files(source_path)

    return files


def read_file(path: Path) -> str:
    """
    Read a file
    """
    with open(str(path)) as f:
        content = f.read()

    content = ''.join(filter(lambda x: x in PRINTABLE, content))

    return content


def concat_and_save(path: PurePath, source_files: List[PythonFile]):
    with open(str(path), 'w') as f:
        for i, source in monit.enum(f"Write {path.name}", source_files):
            f.write(f"# PROJECT: {source.project} FILE: {str(source.relative_path)}\n")
            f.write(read_file(source.path) + "\n")


def get_repos_from_readme(filename: str):
    with open(str(lab.get_data_path() / filename), 'r') as f:
        content = f.read()

    link_pattern = re.compile(r"""
        \[(?P<title>[^\]]*)\] # title
        \((?P<utl>[^\)]*)\) # url
    """, re.VERBOSE)

    res = link_pattern.findall(content)

    github_repos = []
    repo_pattern = re.compile(r'https://github.com/(?P<user>[^/]*)/(?P<repo>[^/#]*)$')
    for title, url in res:
        repos = repo_pattern.findall(url)
        for r in repos:
            github_repos.append((r[0], r[1]))

    return github_repos


def get_awesome_pytorch_readme():
    md = urllib.request.urlopen('https://raw.githubusercontent.com/bharathgs/Awesome-pytorch-list/master/README.md')
    content = md.read()

    with open(str(lab.get_data_path() / 'pytorch_awesome.md'), 'w') as f:
        f.write(str(content))


def download_repo(org: str, repo: str, idx: Optional[int]):
    zip_file = Path(lab.get_data_path() / 'download' / f'{org}_{repo}.zip')

    if zip_file.exists():
        return zip_file

    if idx is not None:
        idx_str = f"{idx:03}: "
    else:
        idx_str = ""

    with monit.section(f"{idx_str} {org}/{repo}") as s:
        try:
            zip = urllib.request.urlopen(f'https://github.com/{org}/{repo}/archive/master.zip')
        except urllib.error.HTTPError as e:
            print(e)
            return
        content = zip.read()

        size = len(content) // 1024
        s.message = f"{size :,}KB"

        with open(str(zip_file), 'wb') as f:
            f.write(content)

    return zip_file


def create_folders():
    path = Path(lab.get_data_path() / 'download')
    if not path.exists():
        path.mkdir(parents=True)
    source = Path(lab.get_data_path() / 'source')

    if not source.exists():
        source.mkdir(parents=True)


def extract_zip(file_path: Path, overwrite: bool = False):
    source = Path(lab.get_data_path() / 'source')

    with monit.section(f"Extract {file_path.stem}"):
        repo_source = source / file_path.stem
        if repo_source.exists():
            if overwrite:
                rm_tree(repo_source)
            else:
                return repo_source
        try:
            with zipfile.ZipFile(file_path, 'r') as repo_zip:
                repo_zip.extractall(repo_source)
        except zipfile.BadZipfile as e:
            print(file_path, e)

        return repo_source


def remove_files(path: Path, keep: Set[str]):
    """
    Remove files
    """

    for p in path.iterdir():
        if p.is_symlink():
            p.unlink()
            continue
        if p.is_dir():
            remove_files(p, keep)
        else:
            if p.suffix not in keep:
                p.unlink()


def batch(overwrite: bool = False):
    with monit.section('Get pytorch_awesome'):
        get_awesome_pytorch_readme()
        repos = get_repos_from_readme('pytorch_awesome.md')

    # Download zips
    for i, r in monit.enum(f"Download {len(repos)} repos", repos):
        download_repo(r[0], r[1], i)

    # Extract downloads
    with monit.section('Extract zips'):
        download = Path(lab.get_data_path() / 'download')

        for repo in download.iterdir():
            extract_zip(repo, overwrite)

    with monit.section('Remove non python files'):
        remove_files(lab.get_data_path() / 'source', {'.py'})


def progressive(overwrite: bool = False):
    # Get repos
    get_awesome_pytorch_readme()
    repos = get_repos_from_readme('pytorch_awesome.md')

    # Download zips
    for i, r in monit.enum(f"Download {len(repos)} repos", repos):
        zip_file = download_repo(r[0], r[1], i)
        extracted = extract_zip(zip_file, overwrite)
        remove_files(extracted, {'.py'})


def main():
    try:
        batch()
    except KeyboardInterrupt:
        pass

    source_files = get_python_files()

    np.random.shuffle(source_files)

    logger.inspect(source_files)

    train_valid_split = int(len(source_files) * 0.9)
    concat_and_save(lab.get_data_path() / 'train.py', source_files[:train_valid_split])
    concat_and_save(lab.get_data_path() / 'valid.py', source_files[train_valid_split:])


if __name__ == '__main__':
    main()



from labml import experiment, lab

if __name__ == '__main__':
    experiment.save_bundle(lab.get_path() / 'bundle.tar.gz', 'a6cff3706ec411ebadd9bf753b33bae6',
                           data_files=['cache/itos.json',
                                       'cache/n_tokens.json',
                                       'cache/stoi.json',
                                       'cache/bpe.json',
                                       ])



from labml import experiment, lab
from labml.utils.pytorch import get_modules
from python_autocomplete.evaluate import Predictor
from python_autocomplete.train import Configs


def load_experiment() -> Configs:
    conf = Configs()
    experiment.evaluate()

    # This will download a pretrained model checkpoint and some cached files.
    # It will download the archive as `saved_checkpoint.tar.gz` and extract it.
    #
    # If you have a locally trained model load it directly with
    # run_uuid = 'RUN_UUID'
    # And for latest checkpoint
    # checkpoint = None

    # run_uuid = 'a6cff3706ec411ebadd9bf753b33bae6'  # bpe
    # checkpoint = None
    run_uuid, checkpoint = experiment.load_bundle(
        lab.get_path() / 'saved_checkpoint.tar.gz',
        url='https://github.com/lab-ml/python_autocomplete/releases/download/0.0.5/bundle.tar.gz')

    conf_dict = experiment.load_configs(run_uuid)
    conf_dict['text.is_load_data'] = False
    conf_dict['device.cuda_device'] = 1
    experiment.configs(conf, conf_dict)
    experiment.add_pytorch_models(get_modules(conf))
    experiment.load(run_uuid, checkpoint)

    experiment.start()

    return conf


def get_predictor() -> Predictor:
    conf = load_experiment()
    conf.model.eval()
    return Predictor(conf.model, conf.text.tokenizer,
                     state_updater=conf.state_updater,
                     is_token_by_token=conf.is_token_by_token)



import torch
from torch import nn

from labml import logger, lab, monit
from labml.logger import Text, Style
from labml_helpers.module import Module
from python_autocomplete.dataset import Tokenizer
from python_autocomplete.evaluate.factory import load_experiment
from python_autocomplete.train import StateUpdater


def anomalies(tokenizer: Tokenizer, text: str, model: Module, state_updater: StateUpdater, is_token_by_token: bool):
    tokens = tokenizer.encode(text)

    line_no = 1
    logs = [(f"{line_no: 4d}: ", Text.meta), (tokenizer.itos[tokens[0]], Style.bold)]

    text = torch.tensor(tokens, dtype=torch.long, device=model.device)
    prompt = text[:1].unsqueeze(-1)

    state = None
    softmax = nn.Softmax(-1)

    i = 1

    while i + 1 < len(text):
        with torch.no_grad():
            prediction, new_state = model(prompt, state)

        state = state_updater(state, new_state)
        prediction = softmax(prediction[-1, 0])

        if is_token_by_token:
            prompt = text[i: i + 1].unsqueeze(-1)
        else:
            prompt = text[:i + 1]
            prompt = prompt[-512:].unsqueeze(-1)

        token_str = tokenizer.itos[text[i]]
        prob = prediction[text[i]].item()

        for c in token_str:
            if c == '\n':
                logger.log(logs)
                line_no += 1
                logs = [(f"{line_no: 4d}: ", Text.meta)]
            elif c == '\r':
                continue
            else:
                if prob > 0.9:
                    logs.append((c, [Text.subtle, Style.underline]))
                elif prob > 0.75:
                    logs.append((c, [Text.success, Style.underline]))
                elif prob > 0.2:
                    logs.append(c)
                elif prob > 0.1:
                    logs.append((c, [Text.warning, Style.underline]))
                elif prob > 0.01:
                    logs.append((c, [Style.bold, Text.warning, Style.underline]))
                elif prob > 0.001:
                    logs.append((c, [Text.danger, Style.underline]))
                else:
                    logs.append((c, [Style.bold, Text.danger, Style.underline]))

        i += 1

    logger.log(logs)


def main():
    conf = load_experiment()

    with open(str(lab.get_data_path() / 'sample.py'), 'r') as f:
        sample = f.read()
    with monit.section('Anomalies'):
        anomalies(conf.text.tokenizer, sample, conf.model, conf.state_updater, conf.is_token_by_token)


if __name__ == '__main__':
    main()



from torch import nn

from labml_nn.recurrent_highway_networks import RHN
from python_autocomplete.models import AutoregressiveModel


class RhnModel(AutoregressiveModel):
    def __init__(self, *,
                 n_tokens: int,
                 embedding_size: int,
                 hidden_size: int,
                 n_layers: int,
                 depth: int):
        super().__init__()

        self.embedding = nn.Embedding(n_tokens, embedding_size)
        self.rhn = RHN(input_size=embedding_size,
                       hidden_size=hidden_size,
                       n_layers=n_layers,
                       depth=depth)
        self.fc = nn.Linear(hidden_size, n_tokens)

    def __call__(self, x, state=None):
        # shape of x is [seq, batch, feat]
        x = self.embedding(x)
        out, s = self.rhn(x, state)
        logits = self.fc(out)

        return logits, s



import torch

from labml import logger, lab, monit
from labml.logger import Text, Style
from python_autocomplete.evaluate import Predictor
from python_autocomplete.evaluate.beam_search import NextWordPredictionComplete
from python_autocomplete.evaluate.factory import get_predictor


def evaluate(predictor: Predictor, text: str):
    line_no = 1
    logs = [(f"{line_no: 4d}: ", Text.meta), (text[0], Text.subtle)]

    correct = 0
    i = 0
    key_strokes = 0

    while i + 1 < len(text):
        prefix = text[:i + 1]
        stripped, prompt = predictor.rstrip(prefix)
        rest = prefix[len(stripped):]
        prediction_complete = NextWordPredictionComplete(rest, 5)
        prompt = torch.tensor(prompt, dtype=torch.long).unsqueeze(-1)

        predictions = predictor.get_next_word(prompt, None, rest, [1.], prediction_complete, 5)
        predictions.sort(key=lambda x: -x[0])
        if predictions:
            next_token = predictions[0].text[len(rest):]
        else:
            next_token = ''

        if next_token and next_token == text[i + 1: i + 1 + len(next_token)]:
            correct += len(next_token)
            right = True
        else:
            next_token = text[i + 1]
            right = False

        for j, c in enumerate(next_token):
            if c == '\n':
                logger.log(logs)
                line_no += 1
                logs = [(f"{line_no: 4d}: ", Text.meta)]
            elif c == '\r':
                continue
            else:
                if right:
                    if j == 0:
                        logs.append((c, [Text.meta, Style.underline]))
                    else:
                        logs.append((c, [Text.success, Style.underline]))
                else:
                    logs.append((c, [Text.warning]))

        i += len(next_token)
        key_strokes += 1

    logger.log(logs)

    logger.inspect(accuracy=correct / (len(text) - 1),
                   key_strokes=key_strokes,
                   length=len(text))


def main():
    predictor = get_predictor()

    with open(str(lab.get_data_path() / 'sample.py'), 'r') as f:
        sample = f.read()
    with monit.section('Evaluate'):
        evaluate(predictor, sample)


if __name__ == '__main__':
    main()



import torch
import torch.nn as nn

from labml import experiment, monit, logger, tracker
from labml.configs import option
from labml.logger import Text
from labml_helpers.device import DeviceConfigs
from labml_helpers.metrics.accuracy import Accuracy
from labml_helpers.metrics.simple_state import SimpleStateModule
from labml_helpers.module import Module
from labml_helpers.train_valid import TrainValidConfigs, hook_model_outputs, BatchIndex
from labml_nn.optimizers.configs import OptimizerConfigs
from labml_nn.transformers import TransformerConfigs
from python_autocomplete.dataset.dataset import SourceCodeDataConfigs


class Configs(TrainValidConfigs):
    optimizer: torch.optim.Adam
    device: torch.device = DeviceConfigs()

    model: Module
    text = SourceCodeDataConfigs()
    n_tokens: int
    n_layers: int = 2
    dropout: float = 0.2
    d_model: int = 512
    rnn_size: int = 512
    rhn_depth: int = 1
    inner_iterations = 100

    is_save_models = True

    transformer: TransformerConfigs

    accuracy = Accuracy()
    loss_func: 'CrossEntropyLoss'

    state_updater: 'StateUpdater'
    state = SimpleStateModule()
    mem_len: int = 512
    grad_norm_clip: float = 1.0
    is_token_by_token: bool = False

    def init(self):
        tracker.set_queue("loss.*", 20, True)
        tracker.set_scalar("accuracy.*", True)
        hook_model_outputs(self.mode, self.model, 'model')
        self.state_modules = [self.accuracy, self.state]

    def step(self, batch: any, batch_idx: BatchIndex):
        data, target = batch[0].to(self.device), batch[1].to(self.device)

        if self.mode.is_train:
            tracker.add_global_step(target.shape[0] * target.shape[1])

        with self.mode.update(is_log_activations=batch_idx.is_last):
            state = self.state.get()
            output, new_state = self.model(data, state)
            state = self.state_updater(state, new_state)
            self.state.set(state)

        loss = self.loss_func(output, target)
        tracker.add("loss.", loss)

        self.accuracy(output, target)
        self.accuracy.track()

        if self.mode.is_train:
            loss.backward()

            torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=self.grad_norm_clip)
            self.optimizer.step()
            if batch_idx.is_last:
                tracker.add('model', self.model)
            self.optimizer.zero_grad()

        tracker.save()

    def sample(self):
        prompt = 'def train('
        log = [(prompt, Text.subtle)]
        state = None
        for i in monit.iterate('Sample', 25):
            data = self.text.text_to_i(prompt).unsqueeze(-1)
            data = data.to(self.device)
            output, new_state = self.model(data, state)
            output = output.argmax(dim=-1).squeeze(1)
            prompt += '' + self.text.tokenizer.itos[output[-1]]
            if self.is_token_by_token:
                prompt = self.text.tokenizer.itos[output[-1]]
            else:
                prompt += '' + self.text.tokenizer.itos[output[-1]]
            log += [('' + self.text.tokenizer.itos[output[-1]], Text.value)]
            state = self.state_updater(state, new_state)

        logger.log(log)


@option(Configs.transformer)
def default_transformer(c: Configs):
    conf = TransformerConfigs()
    conf.d_model = c.d_model
    conf.n_layers = c.n_layers
    conf.n_src_vocab = c.n_tokens
    conf.n_tgt_vocab = c.n_tokens
    conf.dropout = c.dropout

    return conf


@option(Configs.optimizer)
def _optimizer(c: Configs):
    optimizer = OptimizerConfigs()
    optimizer.parameters = c.model.parameters()
    optimizer.optimizer = 'Adam'
    optimizer.d_model = c.d_model

    return optimizer


class CrossEntropyLoss(Module):
    def __init__(self, n_tokens: int):
        super().__init__()
        self.n_tokens = n_tokens
        self.loss = nn.CrossEntropyLoss()

    def __call__(self, outputs, targets):
        return self.loss(outputs.view(-1, self.n_tokens), targets.view(-1))


@option(Configs.loss_func)
def _loss_func(c: Configs):
    return CrossEntropyLoss(c.n_tokens)


@option(Configs.n_tokens)
def _n_tokens(c: Configs):
    return c.text.tokenizer.n_tokens


@option(Configs.model)
def lstm_model(c: Configs):
    from python_autocomplete.models.lstm import LstmModel
    m = LstmModel(n_tokens=c.n_tokens,
                  embedding_size=c.d_model,
                  hidden_size=c.rnn_size,
                  n_layers=c.n_layers)
    return m.to(c.device)


@option(Configs.model)
def rhn_model(c: Configs):
    from python_autocomplete.models.highway import RhnModel
    m = RhnModel(n_tokens=c.n_tokens,
                 embedding_size=c.d_model,
                 hidden_size=c.rnn_size,
                 n_layers=c.n_layers,
                 depth=c.rhn_depth)
    return m.to(c.device)


@option(Configs.model)
def transformer_model(c: Configs):
    from python_autocomplete.models.transformer import TransformerModel
    m = TransformerModel(n_tokens=c.n_tokens,
                         d_model=c.d_model,
                         encoder=c.transformer.encoder,
                         src_embed=c.transformer.src_embed)

    return m.to(c.device)


@option(Configs.model)
def transformer_xl_model(c: Configs):
    from labml_nn.transformers.xl import RelativeMultiHeadAttention
    from labml_nn.transformers.feed_forward import FeedForward
    from labml_nn.transformers.xl import TransformerXL
    from labml_nn.transformers.xl import TransformerXLLayer
    from python_autocomplete.models.xl import TransformerXLModel
    m = TransformerXLModel(c.n_tokens, c.d_model, TransformerXL(
        TransformerXLLayer(d_model=c.d_model,
                           self_attn=RelativeMultiHeadAttention(c.transformer.n_heads, c.d_model, c.dropout),
                           feed_forward=FeedForward(c.d_model, c.transformer.ffn.d_ff, c.dropout),
                           dropout_prob=c.dropout), c.n_layers))
    return m.to(c.device)


class StateUpdater:
    def __call__(self, old_state, new_state):
        return new_state

    def get_from_batch(self, state, batch_idx):
        if state is None:
            return None
        elif isinstance(state, torch.Tensor):
            return state[batch_idx]
        elif isinstance(state, tuple):
            return tuple(s[batch_idx] for s in state)
        elif isinstance(state, list):
            return [s[batch_idx] for s in state]

    def make_batch(self, batch):
        assert isinstance(batch, list)
        if batch[0] is None:
            return None
        elif isinstance(batch[0], torch.Tensor):
            return torch.stack(batch)
        elif isinstance(batch[0], tuple):
            return tuple(torch.stack([b[n] for b in batch]) for n in range(len(batch[0])))
        elif isinstance(batch[0], list):
            return [torch.stack([b[n] for b in batch]) for n in range(len(batch[0]))]


class MemoryUpdater(StateUpdater):
    def __init__(self, mem_len: int):
        self.mem_len = mem_len

    def __call__(self, old_mem, new_mem):
        if self.mem_len == 0:
            return []

        if old_mem:
            mem = [torch.cat((m, x), dim=0) for m, x in zip(old_mem, new_mem)]
        else:
            mem = new_mem

        if len(mem[0]) > self.mem_len:
            mem = [m[-self.mem_len:] for m in mem]

        return mem

    def get_from_batch(self, state, batch_idx):
        if state is None:
            return None

        return [m[:, batch_idx] for m in state]

    def make_batch(self, batch):
        if batch[0] is None:
            return None
        return [torch.stack([b[n] for b in batch], dim=1) for n in range(len(batch[0]))]


@option(Configs.state_updater)
def simple():
    return StateUpdater()


@option(Configs.state_updater)
def transformer_memory(c: Configs):
    return MemoryUpdater(c.mem_len)


@option(Configs.train_loader)
def _train_loader(c: Configs):
    return c.text.train_loader


@option(Configs.valid_loader)
def _valid_loader(c: Configs):
    return c.text.valid_loader


def main():
    conf = Configs()
    # Assign one of transformer_mode, lstm_model, or rhn_model
    experiment.create(name="source_code",
                      comment='bpe')
    experiment.configs(conf, {
        # 'model': 'transformer_model',
        'model': 'transformer_xl_model',
        'n_layers': 6,
        'epochs': 32,
        'optimizer.optimizer': 'AdamW',
        'optimizer.learning_rate': 1.25e-4,
        'device.cuda_device': 0,

        'is_token_by_token': True,
        'state_updater': 'transformer_memory',
        'mem_len': 256,

        'text.is_shuffle': False,
        'text.tokenizer': 'bpe',
        'text.batch_size': 12,
        'text.seq_len': 256,
        #
        # 'inner_iterations': 10,
        # 'text.truncate_data': 100_000,
    })
    experiment.add_pytorch_models(model=conf.model)
    with experiment.start():
        conf.run()


if __name__ == '__main__':
    main()



from typing import List, Tuple

from labml import monit
from python_autocomplete.dataset import ID_CHARS


class WordTokenizer:
    def collect_words(self, data: str):
        raise NotImplementedError

    def get_words(self) -> Tuple[List[str], List[int]]:
        raise NotImplementedError

    def tokenize(self, data: str, *, is_silent: bool = False) -> List[str]:
        raise NotImplementedError


class SourceCodeTokenizer(WordTokenizer):
    def __init__(self):
        self.words = {}

    def add_word(self, word):
        if not word:
            return

        if word not in self.words:
            self.words[word] = 1
        else:
            self.words[word] += 1

    def tokenize(self, data: str, *, is_silent: bool = False) -> List[str]:
        last_idx = 0
        is_id = False
        res = []

        for i, c in monit.enum('Collect words', data, is_silent=is_silent):
            if c in ID_CHARS:
                if not is_id:
                    if last_idx < i:
                        res.append(data[last_idx:i])
                    last_idx = i
                    is_id = True
            else:
                if is_id:
                    if last_idx < i:
                        res.append(data[last_idx:i])
                    last_idx = i
                    is_id = False

        if last_idx < len(data):
            res.append(data[last_idx:])

        return res

    def collect_words(self, data: str):
        last_idx = 0
        is_id = False

        for i, c in monit.enum('Collect words', data):
            if c in ID_CHARS:
                if not is_id:
                    self.add_word(data[last_idx:i])
                    last_idx = i
                    is_id = True
            else:
                if is_id:
                    self.add_word(data[last_idx:i])
                    last_idx = i
                    is_id = False

        self.add_word(data[last_idx:])

    def get_words(self):
        words_list = [(f, w) for w, f in self.words.items()]
        words_list.sort(key=lambda x: -x[0])

        return [w for _, w in words_list], [f for f, _ in words_list]


class NoTokenizer(WordTokenizer):
    def __init__(self):
        self.data = ''

    def collect_words(self, data):
        self.data += data

    def get_words(self):
        return [self.data], [1]

    def tokenize(self, data: str, *, is_silent: bool = False) -> List[str]:
        return [data]



from heapq import heappush, heappop
from typing import Any, List, Optional

import torch
import torch.nn

from python_autocomplete.evaluate.beam_search import PredictionComplete, BeamSearch
from python_autocomplete.train import StateUpdater

EPS_SCORE = 1e-6
MIN_BEAM_PROB = 1e-4


class BeamSearchLengthy(BeamSearch):
    """
    Use a score instead of probability to make longer predictions
    """
    def __init__(self, *, beam_size: int, prediction_complete: PredictionComplete,
                 max_beam_size: int, rest: str,
                 state_updater: 'StateUpdater',
                 probs: Optional[List[float]],
                 is_token_by_token: bool,
                 itos: List[str]):
        super().__init__()
        self.itos = itos
        self.is_token_by_token = is_token_by_token
        self.state_updater = state_updater
        self.prediction_complete = prediction_complete
        self.max_beam_size = max_beam_size
        self.rest = rest

        if probs is None:
            probs = [1 / beam_size] * beam_size
        assert len(probs) == beam_size
        self.probs = probs

        self.result_heap = []
        self.text = [''] * beam_size
        self.beam_heap = []

    @staticmethod
    def is_prefix(original, token_str):
        """
        Whether `token_str` is a prefix of `original`
        """
        if not original:
            return True

        n = min(len(original), len(token_str))
        return original[:n] == token_str[:n]

    @staticmethod
    def _get_score(text: str, prob: float):
        return len(text) ** 0.2 * prob

    def _add_prediction_before_token(self, prob: float, beam_idx: int, state):
        """
        Add a prediction before the last token to the results
        """
        text = self.text[beam_idx]
        score = self._get_score(text, prob)

        if len(self.result_heap) == self.max_beam_size:
            if self.result_heap[0][0] > score - EPS_SCORE:
                return
            heappop(self.result_heap)

        state = self.state_updater.get_from_batch(state, beam_idx)
        heappush(self.result_heap, (score, (text, state)))

    def _add_beam(self, prob: float, beam_idx: int, token: int):
        """Add to the beam"""
        text = self.text[beam_idx] + self.itos[token]
        score = self._get_score(text, prob)

        # if self.result_heap and self.result_heap[0][0] > score - EPS_SCORE:
        #     return False

        if prob < MIN_BEAM_PROB:
            return False

        if len(self.beam_heap) == self.max_beam_size:
            if self.beam_heap[0][0] > score - EPS_SCORE:
                return False
            heappop(self.beam_heap)

        heappush(self.beam_heap, (score, (beam_idx, prob, token)))

        return True

    def next_batch(self, prompt: torch.Tensor, state: Any):
        """Get the next batch (beam)"""
        if not self.beam_heap:
            return None, None

        new_prompt = []
        new_state = []

        texts = self.text
        self.text = []
        self.probs = []

        for _, (b, prob, token) in self.beam_heap:
            token = prompt.new_tensor([token])
            if self.is_token_by_token:
                new_prompt.append(token)
            else:
                new_prompt.append(torch.cat((prompt[1:, b], token)))
            new_state.append(self.state_updater.get_from_batch(state, b))
            self.probs.append(prob)
            self.text.append(texts[b] + self.itos[token])

        new_prompt = torch.stack(new_prompt, dim=1)
        new_state = self.state_updater.make_batch(new_state)

        self.beam_heap = []

        return new_prompt, new_state

    def update(self, next_token, state, old_state):
        """Update beam search with the sampled data"""
        self.beam_heap = []

        for b, text in enumerate(self.text):
            text = self.text[b]
            if len(text) >= len(self.rest):
                check_rest = None
            else:
                check_rest = self.rest[len(text):]

            tokens = next_token[b]
            sort_idx = torch.argsort(tokens)
            added_to_results = False

            for i in reversed(range(len(tokens))):
                token = sort_idx[i]
                token_str = self.itos[token]
                if not self.is_prefix(check_rest, token_str):
                    continue

                if not added_to_results and self.prediction_complete(text, token_str):
                    added_to_results = True
                    self._add_prediction_before_token(self.probs[b], b, old_state)

                if check_rest and len(token_str) <= len(check_rest):
                    p = 1.0
                else:
                    p = tokens[token].item()

                if not self._add_beam(self.probs[b] * p, b, token):
                    break
